



python "%PREFIX%\\Library\\bin\\grcc.py" "%PREFIX%\\Library\\share\\gnuradio\\examples\\metadata\\file_metadata_vector_sink.grc"
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Scripts\\gnuradio-companion.exe exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Scripts\\grcc.exe exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
